/*
 * Copyright (c) 1990 The Regents of the University of California.
 * Copyright (c) 1992 Transarc Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#if 0
static char rcsid[] = "@(#) $Header: /afs/stacken.kth.se/src/SourceRepository/tcpdump/print-afs.c,v 1.2 1998/07/19 18:04:38 mattiasa Exp $";
#endif

#include <stdio.h>
#include <ctype.h>
#include <errno.h>

#include <sys/time.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>

#include <rpc/rpc.h>
#include <rpc/types.h>
#include <rpc/xdr.h>

#include "interfac.h"
#include "a2name.h"
#include "extract.h"
#include "afsops.h"
#include "rx.h"
#include "kasops.h"
#include "vlsops.h"
#include "volops.h"
#include "ubikops.h"
#include "rxkops.h"
#include "ptsops.h"

static char *rxtypes[RX_N_PACKET_TYPES] = RX_PACKET_TYPES;
static char *rxflags[] = RX_FLAGS;
static char *rx_ackReasons = RX_ACK_REASONS;

static char *afsopstrings[81] = {
            "FetchData      ",
            "FetchACL       ",
            "FetchStatus    ",
            "StoreData     ",
            "StoreACL      ",
            "StoreStatus   ",
            "RemoveFile    ",
            "CreateFile    ",
            "Rename                ",
            "Symlink       ",
            "Link          ",
            "MakeDir       ",
            "RemoveDir     ",
            "OldSetLock    ",
            "OldExtendLock ",
            "OldReleaseLock        ",
            "GetStatistics ",
            "GiveUpCallBacks",
            "GetVolumeInfo ",
            "GetVolumeStatus",
            "SetVolumeStatus",
            "GetRootVolume ",
            "CheckToken    ",
            "GetTime       ",
            "NGetVolumeInfo        ",
            "BulkStatus    ",
            "SetLock       ",
            "ExtendLock    ",
            "ReleaseLock   ",
            "XStatsVersion ",
            "GetXStats     ",
            "Lookup         ",
            "FlushCPS       ",
            "DFSSymlink     ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "unknown        ",
            "CallBack      ",
            "InitCallBackState",
            "Probe                 ",
            "GetLock       ",
            "GetCE                 ",
            "XStatsVersion         ",
            "GetXStats     "
          };

static char *afsports[] = {
            "fs", "cm", "pt", "vl", "ka",
            "vs", "er", "bs", "up", "rs",
            "xx", "xx", "xx", "xx", "xx",
            "xx", "xx", "xx", "xx", "xx",
            "xx", "bd"
          };

static char *rxkopstrings[RXKST_NUMBER_OPCODES] = {
            "Fast   ",
            "Slow   ",
            "Copious",
            "Kill   "
          };

static void rxack_print  (struct rx_header*, int, struct ip*, char*, char*);
static void afsreq_print (u_long *, int, struct ip *);
static void kasreq_print (u_long *, int, struct ip *);
static void ubik_print   (u_long *, int, struct ip *);
static void vlsreq_print (u_long *, int, struct ip *);
static void ptsreq_print (u_long *, int, struct ip *);
static void volreq_print (u_long *, int, struct ip *);
static void rxkreq_print (u_long *, int, struct ip *);

static __inline void printname (u_long len, char *dp)
{
  PRINTF (" %8.20s", dp);
}

static __inline void printname2 (u_long len, u_long *dp)
{
  PUTCHAR (' ');
  while (*dp)
  {
    u_long t = ntohl (*dp);
    dp++;
    PUTCHAR (t & 0xff);
  }
}

static __inline void printfid (u_long *dp)
{
  /* try to print out the FID at the beginning of this struct...
   */
  PRINTF (" %ld.%ld.%ld", ntohl(dp[0]), ntohl(dp[1]), ntohl(dp[2]));
}

static __inline void printlong (u_long *dp)
{
  PRINTF (" %ld", ntohl(dp[0]));
}

static __inline void printhost (u_long *dp)
{
  u_long ip = ntohl (dp[0]);

  PRINTF (" %d.%d.%d.%d",
          (ip >> 24) & 0xff,
          (ip >> 16) & 0xff,
          (ip >> 8) & 0xff,
          ip & 0xff);
}

void rx_print (const u_char *p, u_int length, u_int sp, u_int dp,
               u_char *bp2)
{
  struct rx_header *rp = (struct rx_header*) p;
  struct ip        *ip = (struct ip*) bp2;
  char   t[8], u[8];
  char  *sport, *dport;

  if (sp >= AFS_FSPORT && sp <= 7021)
     sport = afsports[sp - SMALLESTPORT];
  else
  {
    sprintf (t, "%x", sp);
    sport = t;
  }

  if (dp >= AFS_FSPORT && dp <= 7021)
     dport = afsports [dp - SMALLESTPORT];
  else
  {
    sprintf (u, "%x", dp);
    dport = u;
  }

  if (length < RX_HEADER_SIZE)  /* must not be truncated, I guess */
  {
    PRINTF ("%s.%s > %s.%s: truncated-RX",
            ipaddr_string (&ip->ip_src), sport,
            ipaddr_string (&ip->ip_dst), dport);
    return;
  }

  if (!nflag)
  {
    if (rp->type == RX_PACKET_TYPE_ACK)
       rxack_print (rp, length, ip, sport, dport);
    else
    {
      if (rp->flags & RX_CLIENT_INITIATED)
      {
        PRINTF ("%s.%s.%08x.%04x > %s.%s: ",
                ipaddr_string (&ip->ip_src), sport,
                ntohl (rp->cid), ntohl (rp->callNumber),
                ipaddr_string (&ip->ip_dst), dport);
        PRINTF ("%2x.%04x %s %s %3d",
                ntohl (rp->seq), ntohl (rp->serial),
                rp->type ? rxtypes[rp->type - 1] : "",
                rxflags[rp->flags], length - RX_HEADER_SIZE);

        if (rp->type == RX_PACKET_TYPE_ABORT)
           PRINTF (" %d", ntohl (*((long*)(rp+1))));

        if (ntohl (rp->seq) == 1)
        {
          if (dp == AFS_FSPORT)
             afsreq_print ((u_long*)(rp+1), length - RX_HEADER_SIZE, ip);
          else if (sp == AFS_FSPORT)
          {
            if (dp == AFS_PTPORT)
                 ptsreq_print ((u_long*)(rp+1), length - RX_HEADER_SIZE, ip);
            else afsreq_print ((u_long*)(rp+1), length - RX_HEADER_SIZE, ip);
          }
          else if (sp == dp)
             ubik_print ((u_long*)(rp+1), length - RX_HEADER_SIZE, ip);
          else if (dp == AFS_KAPORT || dp == AFS_VLPORT || dp == AFS_PTPORT)
          {
            u_long op = ntohl (*(u_long*)(rp + 1));

            if ((op >= UVOTE_LowestOpcode && op <= UVOTE_HighestOpcode) ||
                (op >= UDISK_LowestOpcode && op <= UDISK_HighestOpcode))
                 ubik_print ((u_long*)(rp+1), length - RX_HEADER_SIZE, ip);
            else if (dp == AFS_KAPORT)
                 kasreq_print ((u_long*)(rp+1), length - RX_HEADER_SIZE, ip);
            else if (dp == AFS_VLPORT)
                 vlsreq_print ((u_long*)(rp+1), length - RX_HEADER_SIZE, ip);
            else if (dp == AFS_PTPORT)
                 ptsreq_print ((u_long*)(rp+1), length - RX_HEADER_SIZE, ip);
          }
          else if (dp == AFS_VOSPORT)
               volreq_print ((u_long*)(rp+1), length - RX_HEADER_SIZE, ip);
          else if (dp == RXKST_SERVICEPORT)
               rxkreq_print ((u_long*)(rp+1), length - RX_HEADER_SIZE, ip);
          else PUTCHAR ('\n');
        }
      }
      else
      {
        PRINTF ("%s.%s.%08x.%04x < %s.%s: ",
                ipaddr_string (&ip->ip_dst), dport,
                ntohl (rp->cid), ntohl (rp->callNumber),
                ipaddr_string (&ip->ip_src), sport);
        PRINTF ("%2x.%04x %s %s %3d",
                ntohl (rp->seq), ntohl (rp->serial),
                rp->type ? rxtypes[rp->type - 1] : "",
                rxflags[rp->flags], length - RX_HEADER_SIZE);

        if (rp->type == RX_PACKET_TYPE_ABORT)
           PRINTF (" %ld", ntohl(*((long*)(rp+1))));
      }
    }
  }
}

static void rxack_print (struct rx_header *rp, int len, struct ip *ip,
                         char *sport, char *dport)
{
  struct rx_ackPacket *ap;

  if (rp->flags & RX_CLIENT_INITIATED)
  {
    PRINTF ("%s.%s.%08x.%04x > %s.%s: ",
            ipaddr_string (&ip->ip_src),
            sport, ntohl (rp->cid), ntohl (rp->callNumber),
            ipaddr_string (&ip->ip_dst), dport);
  }
  else
  {
    PRINTF ("%s.%s.%08x.%04x < %s.%s: ",
            ipaddr_string (&ip->ip_dst), dport,
            ntohl (rp->cid), ntohl (rp->callNumber),
            ipaddr_string (&ip->ip_src), sport);
  }

  PRINTF ("%2x.", rp->seq);

  rp++;
  ap = (struct rx_ackPacket*) rp;

  if (&ap->nAcks > snapend)
  {
    PUTS ("00 ack-truncated ");
    return;
  }

  PRINTF ("%04x ack %c %x",
          ntohl (ap->serial), rx_ackReasons[(int)ap->reason],
          ntohl (ap->firstPacket));

  if (ap->nAcks)
  {
    int offset;

    for (offset = 0; offset < ap->nAcks; offset++)
    {
      if (&(ap->acks[offset]) > snapend)
           PRINTF ("+%dacks ", ap->nAcks - offset);
      else PUTCHAR (ap->acks[offset] == RX_ACK_TYPE_NACK ? '-' : '*');
    }
  }
}

static void afsreq_print (u_long *rp, int length, struct ip *ip)
{
  u_long op = ntohl (*rp);

  if (op >= AFSOP_LowestOpcode && op <= AFSOP_HighestOpcode)
     PRINTF (" %s", (afsopstrings[op - AFSOP_LowestOpcode]));

  switch (op)
  {
    case AFSOP_FetchStatus:
    case AFSOP_StoreStatus:
    case AFSOP_SetLock:
    case AFSOP_ExtendLock:
    case AFSOP_ReleaseLock:
    case AFSOP_FetchACL:
    case AFSOP_StoreACL:
         printfid (rp+1);
         break;
    case AFSOP_CallBack:
         printfid (rp+2);
         break;
    case AFSOP_RemoveFile:  /* the rest of these guys fids are for parent */
    case AFSOP_CreateFile:
    case AFSOP_Rename:
    case AFSOP_Symlink:
    case AFSOP_Link:
    case AFSOP_MakeDir:
    case AFSOP_RemoveDir:
    case AFSOP_Lookup:
    case AFSOP_DFSSymlink:
         printfid (rp+1);
         printname (*(rp+4), (char*)(rp+5));
         break;
    case AFSOP_FetchData:   /* also print out pos and len */
         printfid (rp+1);
         printlong (rp+4);
         printlong (rp+5);
         break;
    case AFSOP_StoreData:   /* fid, pos, len, file len */
         printfid (rp + 1);
         printlong (rp+10);
         printlong (rp+11);
         printlong (rp+12);
         break;
    case AFSOP_GetVolumeInfo:
         printname (*(rp+1), (char*)(rp+2));
         break;
  }
}

static void kasreq_print (u_long *rp, int length, struct ip *ip)
{
  u_long op = ntohl (*rp);

  if (op >= KASOP_LowestOpcode && op <= KASOP_HighestOpcode)
     PRINTF (" %s", (kasopstrings[op - KASOP_LowestOpcode]));

  switch (op)
  {
    case KASOP_Authenticate_old:
    case KASOP_Authenticate:
    case KASOP_ChangePassword:
    case KASOP_SetPassword:
    case KASOP_SetFields:
    case KASOP_CreateUser:
    case KASOP_DeleteUser:
    case KASOP_GetEntry:
    case KASOP_GetPassword:
         printname (*(rp+1), (char*)(rp+2));
  }
}

static void ubik_print (u_long *rp, int length, struct ip *ip)
{
  u_long op = ntohl (*rp);

  if (op >= UVOTE_LowestOpcode && op <= UVOTE_HighestOpcode)
       PRINTF (" %s", (UVOTE_opstrings[op - UVOTE_LowestOpcode]));
  else if (op >= UDISK_LowestOpcode && op <= UDISK_HighestOpcode)
       PRINTF (" %s", (UDISK_opstrings[op - UDISK_LowestOpcode]));

#if 0
  switch (op)
  {
    case VOTE_BEACON:
    case VOTE_DEBUG:
    case VOTE_SDEBUG:
    case VOTE_GETSYNCSITE:
    case DISK_BEGIN:
    case DISK_COMMIT:
    case DISK_LOCK:
    case DISK_WRITE:
    case DISK_GETVERSION:
    case DISK_GETFILE:
    case DISK_SENDFILE:
    case DISK_ABORT:
    case DISK_RELEASELOCKS:
    case DISK_TRUNCATE:
    case DISK_PROBE:
         break;
  }
#endif
}

static void vlsreq_print (u_long *rp, int length, struct ip *ip)
{
  u_long op = ntohl (*rp);

  if (op >= VLSOP_LowestOpcode && op <= VLSOP_HighestOpcode)
     PRINTF (" %s", (vlsopstrings[op - VLSOP_LowestOpcode]));

  switch (op)
  {
    case VLSOP_GetEntryByID:
    case VLGETENTRYBYIDN:
    case VLDELETEENTRY:
    case VLREPLACEENTRY:
    case VLREPLACEENTRYN:
    case VLUPDATEENTRY:
    case VLSETLOCK:
    case VLRELEASELOCK:
         printlong (rp+1);
         break;
    case VLUPDATEENTRYBYNAME:
    case VLGETENTRYBYNAMEN:
    case VLSOP_GetEntryByName:
         printname (*(rp+1), (char*)(rp+2));
         break;
    case VLCHANGEADDR:
         printlong (rp+1);
         printlong (rp+2);
         break;
    case VLCREATEENTRY:
    case VLCREATEENTRYN:
    case VLGETNEWVOLUMEID:
    case VLLISTENTRY:
    case VLLISTENTRYN:
    case VLLISTATTRIBUTES:
    case VLLISTATTRIBUTESN:
    case VLLINKEDLIST:
    case VLLINKEDLISTN:
    case VLGETSTATS:
    case VLPROBE:
         break;
  }
}

static void ptsreq_print (u_long *rp, int length, struct ip *ip)
{
  u_long op = ntohl (*rp);

  if (op >= PTSOP_LowestOpcode && op <= PTSOP_HighestOpcode)
     PRINTF (" %s", (ptsopstrings[op - PTSOP_LowestOpcode]));

  switch (op)
  {
    case PRINEWUSER:
         printname (*(rp+1), (char*)(rp+2));
         break;
    case PRWHEREISIT:
    case PRDUMPENTRY:
         printlong (rp+1);
         break;
    case PRADDTOGROUP:
         printlong (rp+1);
         printlong (rp+2);
         break;
    case PRNAMETOID:
         printname2 (*(rp+2), (u_long*)(rp+3));
         break;
    case PRIDTONAME:
         printlong (rp+1);
         printlong (rp+2);
         break;
    case PRDELETE:
         printlong (rp+1);
         break;
    case PRREMOVEFROMGROUP:
         printlong (rp+1);
         printlong (rp+2);
         break;
    case PRGETCPS:
         printlong (rp+1);
         break;
    case PRNEWENTRY:
         printname (*(rp+1), (char*)(rp+2));
         break;
    case PRLISTMAX:
    case PRSETMAX:
         printlong (rp+1);
         printlong (rp+2);
         break;
    case PRLISTENTRY:
         printname (*(rp+1), (char*)(rp+2));
         break;
    case PRCHANGEENTRY:
         printlong (rp+1);
         printname (*(rp+2), (char*)(rp+3));
         break;
    case PRLISTELEMENTS:
         printlong (rp+1);
         break;
    case PRISAMEMBEROF:
         printlong (rp+1);
         printlong (rp+2);
         break;
    case PRSETFIELDSENTRY:
         printlong (rp+1);
         printlong (rp+2);
         break;
    case PRLISTOWNED:
         printlong (rp+1);
         break;
    case PRGETCPS2:
         printlong (rp+1);
         printlong (rp+2);
         break;
    case PRGETHOSTCPS:
         printhost (rp+1);
         break;
  }
}

static void volreq_print (u_long *rp, int length, struct ip *ip)
{
  u_long op = ntohl (*rp);

  if (op >= VOLOP_LowestOpcode && op <= VOLOP_HighestOpcode)
     PRINTF (" %s", (volopstrings[op - VOLOP_LowestOpcode]));

  switch (op)
  {
    case VOLDELETEVOLUME:
    case VOLRESTORE:
    case VOLFORWARD:
    case VOLENDTRANS:
    case VOLCLONE:
    case VOLSETFLAGS:
    case VOLGETFLAGS:
    case VOLDUMP:
    case VOLSETFORWARDING:
    case VOLGETNAME:
    case VOLGETSTATUS:
    case VOLSETIDSTYPES:
    case VOLRECLONE:
         printlong (rp+1);
         break;
    case VOLNUKE:
    case VOLSETDATE:
    case VOLXLISTVOLS:
    case VOLXLISTONEVOL:
    case VOLSETINFO:
    case VOLXLISTPARTITIONS:
    case VOLFORWARDMULTIPLE:
    case VOLTRANSCREATE:
    case VOLCREATEVOLUME:
    case VOLGETNTHVOLUME:
    case VOLSIGRESTORE:
    case VOLLISTPARTITIONS:
    case VOLLISTONEVOL:
    case VOLLISTVOLS:
    case VOLDISKPART:
    case VOLMONITOR:
         break;
  }
}

static void rxkreq_print (u_long *rp, int length, struct ip *ip)
{
  u_long op = ntohl (*rp);

  if (op >= RXKST_LOWEST_OPCODE && op <= RXKST_HIGHEST_OPCODE)
     PRINTF (" %s", (rxkopstrings[op - RXKST_LOWEST_OPCODE]));

  switch (op)
  {
    case RXKST_Fast:
    case RXKST_Slow:
         printlong (rp+1);
         break;
    case RXKST_Copious:
         printlong (rp+1);
         printlong (rp+2);
         printlong (rp+3);
         break;
    case RXKST_Kill:
         break;
  }
}
