/*      $NetBSD: print-ah.c,v 1.4 1996/05/20 00:41:16 fvdl Exp $        */

/*
 * Copyright (c) 1988, 1989, 1990, 1991, 1992, 1993, 1994
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#if 0
static const char rcsid[] =
"@(#) $Header: /tcpdump/master/tcpdump/print-ah.c,v 1.5 1999/12/15 08:10:17 fenner Exp $ (LBL)";
#endif

#include <stdio.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <net/route.h>
#include <net/if.h>

#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/ip_var.h>
#include <netinet/udp.h>
#include <netinet/udp_var.h>
#include <netinet/tcp.h>


/* there's no standard definition so we are on our own
 */
struct ah
{
  u_int8_t ah_nxt;              /* Next Header */
  u_int8_t ah_len;              /* Length of data, in 32bit */
  u_int16_t ah_reserve;          /* Reserved for future use */
  u_int32_t ah_spi;              /* Security parameter index */
/* variable size, 32bit bound *//* Authentication data */
};

struct newah
{
  u_int8_t ah_nxt;              /* Next Header */
  u_int8_t ah_len;              /* Length of data + 1, in 32bit */
  u_int16_t ah_reserve;          /* Reserved for future use */
  u_int32_t ah_spi;              /* Security parameter index */
  u_int32_t ah_seq;              /* Sequence number field */
/* variable size, 32bit bound *//* Authentication data */
};

#include "interfac.h"
#include "a2name.h"

int ah_print (const u_char * bp)
{
  const struct ah *ah;
  const u_char    *ep;
  int          sumlen;
  u_int32_t    spi;

  ah = (struct ah *) bp;
  ep = snapend;          /* 'ep' points to the end of avaible data. */

  if ((u_char*)(ah + 1) >= ep - sizeof(struct ah))
     goto trunc;

  sumlen = ah->ah_len << 2;
  spi    = ntohl (ah->ah_spi);

  PRINTF ("AH(spi=%u", spi);
  if (vflag)
     PRINTF (",sumlen=%d", sumlen);

  PRINTF (",seq=0x%x", ntohl(*(u_int32_t*)(ah + 1)));
  if (bp + sizeof(struct ah) + sumlen > ep)
     PUTS ("[truncated]");

  PUTS ("): ");
  return (sizeof(struct ah) + sumlen);

trunc:
  PUTS ("[|AH]");
  return (65535);
}
