/*      $NetBSD: print-ascii.c,v 1.1 1999/09/30 14:49:12 sjg Exp $      */

/*-
 * Copyright (c) 1997, 1998 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Alan Barrett and Simon J. Gerraty.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *        This product includes software developed by the NetBSD
 *        Foundation, Inc. and its contributors.
 * 4. Neither the name of The NetBSD Foundation nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#if 0
static const char rcsid[] =
"@(#) $Header: /tcpdump/master/tcpdump/print-ascii.c,v 1.2.2.2 2000/01/11 06:58:23 fenner Exp $";
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <ctype.h>

#include "interfac.h"

#ifdef TEST_PROG
  #undef  PRINTF
  #undef  PUTS
  #undef  PUTCHAR
  #define PRINTF  printf
  #define PUTS    puts
  #define PUTCHAR putchar
#endif

#define HEXDUMP_BYTES_PER_LINE      16
#define HEXDUMP_SHORTS_PER_LINE     (HEXDUMP_BYTES_PER_LINE / 2)
#define HEXDUMP_HEXSTUFF_PER_SHORT  5  /* 4 hex digits and a space */
#define HEXDUMP_HEXSTUFF_PER_LINE   (HEXDUMP_HEXSTUFF_PER_SHORT * HEXDUMP_SHORTS_PER_LINE)

void ascii_print_with_offset (const u_char * cp, u_int length, u_int oset)
{
  u_int i;
  int   s1, s2, nshorts;
  char  hexstuff [HEXDUMP_SHORTS_PER_LINE * HEXDUMP_HEXSTUFF_PER_SHORT + 1];
  char *hsp;
  char  asciistuff[HEXDUMP_BYTES_PER_LINE + 1];
  char *asp;

  nshorts = length / sizeof(u_short);
  i = 0;
  hsp = hexstuff;
  asp = asciistuff;
  while (--nshorts >= 0)
  {
    s1 = *cp++;
    s2 = *cp++;
    sprintf (hsp, " %02x%02x", s1, s2);
    hsp += HEXDUMP_HEXSTUFF_PER_SHORT;
    *(asp++) = (isgraph (s1) ? s1 : '.');
    *(asp++) = (isgraph (s2) ? s2 : '.');
    if (++i >= HEXDUMP_SHORTS_PER_LINE)
    {
      *hsp = *asp = '\0';
      PRINTF ("\n0x%04x\t%-*s\t%s",
              oset, HEXDUMP_HEXSTUFF_PER_LINE, hexstuff, asciistuff);
      i = 0;
      hsp = hexstuff;
      asp = asciistuff;
      oset += HEXDUMP_BYTES_PER_LINE;
    }
  }
  if (length & 1)
  {
    s1 = *cp++;
    sprintf (hsp, " %02x", s1);
    hsp += 3;
    *(asp++) = (isgraph (s1) ? s1 : '.');
    ++i;
  }
  if (i > 0)
  {
    *hsp = *asp = '\0';
    PRINTF ("\n0x%04x\t%-*s\t%s",
            oset, HEXDUMP_HEXSTUFF_PER_LINE, hexstuff, asciistuff);
  }
}

void ascii_print (const u_char * cp, u_int length)
{
  ascii_print_with_offset (cp, length, 0);
}

#if 0
static char *nice_char_print (char c)
{
  static char buf[3];

  if (isprint(c))
  {
    buf[0] = ' ';
    buf[1] = c;
    goto ret;
  }

  if (iscntrl(c))
  {
    buf[0] = '^';
    buf[1] = '@' + (char)c;
    goto ret;
  }

  if (c == 127)
  {
    buf[0] = '^';
    buf[1] = '?';
    goto ret;
  }

  buf[0] = '.';
  buf[1] = '.';

ret:
  buf[2] = 0;
  return (buf);
}

/*
 * Print buffer out as ASCII characters.  Alignment is not an issue.
 */
void ascii_print (const u_char *bp, u_int length)
{
  int i = 0;
  int even = (length & 1) ? length-1 : length;

  PUTCHAR ('\n');
  while (i < even)
  {
    if ((i % 16) == 0)
       PUTS ("\n\t\t\t");
    PUTCHAR (' ');
    PUTS (nice_char_print(*bp));
    PUTS (nice_char_print(*(bp+1)));
    i += 2;
    bp += 2;
  }
  if (length & 1)
  {
    if ((i % 8) == 0)
       PUTS ("\n\t\t\t");
    PUTCHAR (' ');
    PUTS (nice_char_print(*bp));
  }
}
#endif

/*
 * telnet_print() wants this.  It is essentially default_print_unaligned()
 */
void hex_print_with_offset (const u_char * cp, u_int length, u_int oset)
{
  u_int i, s;
  int nshorts;

  nshorts = (u_int)length / sizeof(u_short);
  i = 0;
  while (--nshorts >= 0)
  {
    if ((i++ % 8) == 0)
    {
      PRINTF ("\n0x%04x\t", oset);
      oset += HEXDUMP_BYTES_PER_LINE;
    }
    s = *cp++;
    PRINTF (" %02x%02x", s, *cp++);
  }
  if (length & 1)
  {
    if ((i % 8) == 0)
       PRINTF ("\n0x%04x\t", oset);
    PRINTF (" %02x", *cp);
  }
}

/*
 * just for completeness
 */
void hex_print (const u_char * cp, u_int length)
{
  hex_print_with_offset (cp, length, 0);
}

#ifdef TEST_PROG

int main (int argc, char *argv[])
{
  hex_print ("Hello, World!\n", 14);
  printf ("\n");
  ascii_print ("Hello, World!\n", 14);
  printf ("\n");

#define TMSG "Now is the winter of our discontent...\n"
  ascii_print_with_offset (TMSG, sizeof(TMSG) - 1, 0x100);
  printf ("\n");
  exit (0);
}
#endif
