/*
 * Copyright (c) 1990, 1991, 1993, 1994, 1995, 1996, 1997
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#if 0
static const char rcsid[] =
"@(#) $Header: /tcpdump/master/tcpdump/print-chdlc.c,v 1.2 1999/11/21 09:36:49 fenner Exp $ (LBL)";
#endif

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include <sys/param.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/ioctl.h>

#include <net/if.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/if_ether.h>
#include <netdb.h>
#include <pcap.h>

#include "ethertyp.h"
#include "interfac.h"
#include "a2name.h"
#include "ppp.h"

/* XXX This goes somewhere else.
 */
#define CHDLC_HDRLEN     4
#define CHDLC_UNICAST    0x0f
#define CHDLC_BCAST      0x8f
#define CHDLC_TYPE_SLARP 0x8035
#define CHDLC_TYPE_CDP   0x2000

static void chdlc_slarp_print (const u_char *, u_int);

/* Standard CHDLC printer
 */
void chdlc_if_print (u_char *user, const struct pcap_pkthdr *h, const u_char * p)
{
  u_int length = h->len;
  u_int caplen = h->caplen;
  const struct ip *ip;
  u_int proto;

  ts_print (&h->ts);

  if (caplen < CHDLC_HDRLEN)
  {
    PUTS ("[|chdlc]");
    goto out;
  }

  /*
   * Some printers want to get back at the link level addresses,
   * and/or check that they're not walking off the end of the packet.
   * Rather than pass them all the way down, we set these globals.
   */
  proto = ntohs (*(u_short *) & p[2]);
  packetp = p;
  snapend = p + caplen;

  if (eflag)
  {
    switch (p[0])
    {
      case CHDLC_UNICAST:
           PUTS ("unicast ");
           break;
      case CHDLC_BCAST:
           PUTS ("bcast ");
           break;
      default:
           PRINTF ("0x%02x ", p[0]);
           break;
    }
    PRINTF ("%d %04x: ", length, proto);
  }

  length -= CHDLC_HDRLEN;
  ip = (struct ip *) (p + CHDLC_HDRLEN);
  switch (proto)
  {
    case ETHERTYPE_IP:
         ip_print ((const u_char *) ip, length);
         break;
#ifdef USE_INET6
    case ETHERTYPE_IPV6:
         ip6_print ((const u_char *) ip, length);
         break;
#endif
    case CHDLC_TYPE_SLARP:
         chdlc_slarp_print ((const u_char *) ip, length);
         break;
#if 0
    case CHDLC_TYPE_CDP:
         chdlc_cdp_print ((const u_char *) ip, length);
         break;
#endif
  }
  if (xflag)
     default_print ((const u_char *) ip, caplen - CHDLC_HDRLEN);
out:
  PUTCHAR ('\n');
}

struct cisco_slarp {
       long code;
#define SLARP_REQUEST   0
#define SLARP_REPLY     1
#define SLARP_KEEPALIVE 2
       union {
         struct {
           struct in_addr addr;
           struct in_addr mask;
           u_short        unused[3];
         } addr;
         struct {
           long  myseq;
           long  yourseq;
           short rel;
           short t1;
           short t2;
         } keep;
       } un;
     };

#define SLARP_LEN  18

static void chdlc_slarp_print (const u_char * cp, u_int length)
{
  struct cisco_slarp *slarp;

  if (length < SLARP_LEN)
  {
    PUTS ("[|slarp]");
    return;
  }

  slarp = (struct cisco_slarp *) cp;
  switch (ntohl (slarp->code))
  {
    case SLARP_REQUEST:
         PUTS ("slarp-request");
         break;
    case SLARP_REPLY:
         PRINTF ("slarp-reply %s/%s",
                 ipaddr_string (&slarp->un.addr.addr),
                 ipaddr_string (&slarp->un.addr.mask));
         break;
    case SLARP_KEEPALIVE:
         PRINTF ("slarp-keepalive my=0x%lx your=0x%lx ",
                 ntohl (slarp->un.keep.myseq),
                 ntohl (slarp->un.keep.yourseq));
         PRINTF ("reliability=0x%04x t1=%d.%d",
                 ntohs (slarp->un.keep.rel),
                 ntohs (slarp->un.keep.t1),
                 ntohs (slarp->un.keep.t2));
         break;
    default:
         PRINTF ("slarp-0lx%x unknown", ntohl (slarp->code));
         break;
  }

  if (SLARP_LEN < length && vflag)
     PRINTF ("(trailing junk: %d bytes)", length - SLARP_LEN);
}
