/*
 * Copyright (c) 1992, 1993, 1994, 1995, 1996
 * The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#if 0
static const char rcsid[] =
  "@(#) $Header: /ng/src/proto-analyser/tcpdump/RCS/print-giop.c,v 1.7 1998/06/28 03:12:43 janssen Exp $ (LBL)";
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <ctype.h>

#include <sys/param.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>
#include <netinet/tcp.h>
#include <netinet/tcpip.h>

#include "interfac.h"
#include "a2name.h"

#define IIOP_TAG_INTERNET_IOP		0
#define IIOP_TAG_MULTIPLE_COMPONENT_IOP	1
#define IIOP_TAG_ILU_IOP		0x494c5500
#define IIOP_TAG_ILU_SUNRPC_IOP		0x494c5501
#define IIOP_TAG_ILU_COURIER_IOP	0x494c5502
#define IIOP_TAG_ILU_UNUSED1_IOP	0x494c5503
#define IIOP_TAG_ILU_UNUSED2_IOP	0x494c5504
#define IIOP_TAG_ILU_UNUSED3_IOP	0x494c5505
#define IIOP_TAG_ILU_UNUSED4_IOP	0x494c5506
#define IIOP_TAG_ILU_UNUSED5_IOP	0x494c5507
#define IIOP_TAG_CODE_SETS		1

#define IIOP_CHARSET_UNICODE_1_1	0x00010100
#define IIOP_CHARSET_UNICODE_UTF_8	0x05010001
#define IIOP_CHARSET_ISO_LATIN1		0x00010001
#define IIOP_CHARSET_US_ASCII		0x00010020

#define GIOP_MSGTYPE_REQUEST		0
#define GIOP_MSGTYPE_REPLY		1
#define GIOP_MSGTYPE_CANCEL_REQUEST	2
#define GIOP_MSGTYPE_LOCATE_REQUEST	3
#define GIOP_MSGTYPE_LOCATE_REPLY	4
#define GIOP_MSGTYPE_CLOSE_CONNECTION	5
#define GIOP_MSGTYPE_MESSAGE_ERROR	6
#define GIOP_MSGTYPE_FRAGMENT		7

#define GIOP_REPLYSTATUS_NO_EXCEPTION		0
#define GIOP_REPLYSTATUS_USER_EXCEPTION		1
#define GIOP_REPLYSTATUS_SYSTEM_EXCEPTION	2
#define GIOP_REPLYSTATUS_LOCATION_FORWARD	3

#define GIOP_SERVICECONTEXT_TRANSACTION_SERVICE	0
#define GIOP_SERVICECONTEXT_CODESETS		1

#define GIOP_PROFILE_IIOP			0
#define GIOP_PROFILE_MULTICOMPONENT		1
#define GIOP_PROFILE_ILU_IOP		0x494c5500
#define GIOP_PROFILE_ILU_SUNRPC_IOP	0x494c5501
#define GIOP_PROFILE_ILU_COURIER_IOP	0x494c5502
#define GIOP_PROFILE_ILU_UNUSED1_IOP	0x494c5503
#define GIOP_PROFILE_ILU_UNUSED2_IOP	0x494c5504
#define GIOP_PROFILE_ILU_UNUSED3_IOP	0x494c5505
#define GIOP_PROFILE_ILU_UNUSED4_IOP	0x494c5506
#define GIOP_PROFILE_ILU_UNUSED5_IOP	0x494c5507

#define IIOP_UNKNOWN_OBJECT			0
#define IIOP_OBJECT_HERE			1
#define IIOP_OBJECT_FORWARD			2

#define IOP_COMPONENTTAG_ORB_TYPE		6
#define IOP_COMPONENTTAG_CODE_SETS		1
#define IOP_COMPONENTTAG_SEC_NAME		14
#define IOP_COMPONENTTAG_ASSOCIATION_OPTIONS	13
#define IOP_COMPONENTTAG_GENERIC_SEC_MECH	12

struct giop_state {
       u_int id;            /* connection id */
     };

struct state_list {
       struct state_list *next;
       struct giop_state  state;
     };

static struct state_list *connections = NULL;

static __inline struct giop_state *add_state (u_int id)
{
  struct state_list *newstate = malloc (sizeof(*newstate));

  newstate->state.id = id;
  newstate->next     = connections;
  connections = newstate;
  return (&newstate->state);
}

static __inline struct giop_state *find_state (u_int id)
{
  struct state_list *p;

  for (p = connections; p; p = p->next)
      if (p->state.id == id)
         return (&p->state);
  return (NULL);
}

static char *quotechars (u_char * p, u_int len)
{
  char  *n = malloc ((4 * len) + 1);
  u_int  i, j;

  if (!n)
     return (NULL);

  for (i = 0, j = 0; i < len; i++)
  {
    if (!isgraph ((int) (p[i])) || (p[i] == '\\'))
    {
      sprintf (n + j, "\\%3.3o", (int) p[i]);
      j += strlen (n + j);
    }
    else
      n[j++] = (char) p[i];
  }
  n[j] = 0;
  return (n);
}

static char *figure_charset_name (u_int32_t charset_code)
{
  switch (charset_code)
  {
    case IIOP_CHARSET_US_ASCII:
         return ("(US ASCII)");
    case IIOP_CHARSET_ISO_LATIN1:
         return ("(ISO Latin1)");
    case IIOP_CHARSET_UNICODE_UTF_8:
         return ("(Unicode UTF-8)");
    case IIOP_CHARSET_UNICODE_1_1:
         return ("(Unicode UCS-2)");
    default:
         return ("");
  }
}

#define SWAP_WORD(a) ( ((a) << 24) |              \
                      (((a) << 8) & 0x00ff0000) | \
                      (((a) >> 8) & 0x0000ff00) | \
		      ((u_int32_t)(a) >>24) )

static u_int32_t gtohl (u_int32_t val, u_int little_endian)
{
  if (!little_endian)
     return SWAP_WORD (val);
  return (val);
}

static u_int print_ior (u_char * ior, u_int little_endian)
{
  u_int32_t typeid_len, nprofiles, profile_len, profile_tag;
  u_char   *typeid, *profile_bytes, *orig_ior;
  char *    s;
  u_int     i;

  orig_ior = ior;

  typeid_len = gtohl (*(u_int32_t*)ior, little_endian);
  ior += 4;
  typeid = ior;
  ior += ((typeid_len + 3) / 4) * 4;
  nprofiles = gtohl (*(u_int32_t*)ior, little_endian);
  ior += 4;

  if (typeid_len > 0)
  {
    s = quotechars (typeid, typeid_len - 1);
    PRINTF ("type=\"%s\"", s);
    free (s);
  }

  for (i = 0; i < nprofiles; i++)
  {
    profile_tag = gtohl (*(u_int32_t*)ior, little_endian);
    ior += 4;
    profile_len = gtohl (*(u_int32_t*)ior, little_endian);
    ior += 4;
    profile_bytes = ior;
    PUTS (",profile(");

    switch (profile_tag)
    {
      case GIOP_PROFILE_IIOP:
           {
             u_int32_t little_endian, major, minor, hostnamelen, port;
             u_int32_t okeylen, ncomponents, component_tag, j, component_len;
             u_int     component_little_endian;
             u_char   *hostname, *okey;
             char     *s;

             little_endian = (profile_bytes[0] != 0);
             major = profile_bytes[1];
             minor = profile_bytes[2];
             profile_bytes += 4;
             hostnamelen    = gtohl (*(u_int32_t*)profile_bytes, little_endian);
             profile_bytes += 4;
             hostname       = profile_bytes;
             profile_bytes += ((hostnamelen + 1) / 2) * 2;
             if (little_endian)
                  port = profile_bytes[0] + (profile_bytes[1] << 8);
             else port = profile_bytes[1] + (profile_bytes[0] << 8);

             profile_bytes += (((((u_int32_t)profile_bytes) % 4) != 0) ? 4 : 2);
             okeylen        = gtohl (*(u_int32_t*)profile_bytes, little_endian);
             profile_bytes += 4;
             okey = profile_bytes;
             s = quotechars (okey, okeylen);

             PRINTF ("type=IIOP,hostname=\"%s\",port=%lu,objectkey=\"%s\"",
                     hostname, port, s);
             free (s);

             if (minor > 0)
             {
               profile_bytes += ((okeylen + 3) / 4) * 4;
               ncomponents    = gtohl (*(u_int32_t*)profile_bytes, little_endian);
               profile_bytes += 4;
               for (j = 0; j < ncomponents; j += 1)
               {
                 component_tag  = gtohl (*(u_int32_t*)profile_bytes, little_endian);
                 profile_bytes += 4;
                 component_len  = gtohl (*(u_int32_t*)profile_bytes, little_endian);
                 profile_bytes += 4;

                 switch (component_tag)
                 {
                   case IOP_COMPONENTTAG_ORB_TYPE:
                        {
                          u_int32_t orb_type;

                          component_little_endian = (profile_bytes[0] != 0);
                          profile_bytes += 4;
                          orb_type = gtohl (*(u_int32_t*)profile_bytes,
                                            component_little_endian);
                          profile_bytes += 4;
                          PRINTF (",orbtype=%lu");
                        }
                        break;

                   case IOP_COMPONENTTAG_CODE_SETS:
                        {
                          u_int32_t ncodesets, j, codeset;

                          component_little_endian = (profile_bytes[0] != 0);
                          profile_bytes += 4;
                          codeset = gtohl (*(u_int32_t*)profile_bytes,
                                           component_little_endian);
                          profile_bytes += 4;
                          PRINTF (",char_codeset=%x%s",
                                  codeset, figure_charset_name (codeset));
                          ncodesets = gtohl (*(u_int32_t*)profile_bytes,
                                             component_little_endian);
                          profile_bytes += 4;
                          for (j = 0; j < ncodesets; j++)
                          {
                            codeset = gtohl (*(u_int32_t*)profile_bytes,
                                             component_little_endian);
                            profile_bytes += 4;
                          }
                          codeset = gtohl (*(u_int32_t*)profile_bytes,
                                           component_little_endian);
                          profile_bytes += 4;
                          PRINTF (",wchar_codeset=%x%s",
                                  codeset, figure_charset_name (codeset));
                          ncodesets = gtohl (*(u_int32_t*)profile_bytes,
                                             component_little_endian);
                          profile_bytes += 4;
                          for (j = 0; j < ncodesets; j++)
                          {
                            codeset = gtohl (*(u_int32_t*)profile_bytes,
                                             component_little_endian);
                            profile_bytes += 4;
                          }
                        }
                        break;

                   case IOP_COMPONENTTAG_SEC_NAME:
                        PRINTF (",component(SEC_NAME,bytes=%lu)",
                                component_len);
                        break;

                   case IOP_COMPONENTTAG_ASSOCIATION_OPTIONS:
                        PRINTF (",component(ASSOCIATION_OPTIONS,bytes=%lu)",
                                component_len);
                        break;

                   case IOP_COMPONENTTAG_GENERIC_SEC_MECH:
                        PRINTF (",component(GENERIC_SEC_MECH,bytes=%lu)",
                                component_len);
                        break;

                   default:
                        PRINTF (",component(tag=%lu,bytes=%lu)",
                                component_tag, component_len);
                        break;
                 }
               }
             }
           }
           break;

      case GIOP_PROFILE_MULTICOMPONENT:
           PRINTF ("type=MultiComponent,bytes=%lu", profile_len);
           break;

      case GIOP_PROFILE_ILU_IOP:
           {
             u_int32_t little_endian, sbhlen;

             little_endian  = (profile_bytes[0] != 0);
             profile_bytes += 4;
             sbhlen         = gtohl (*(u_int32_t*)profile_bytes,
                                     little_endian) - 1;
             profile_bytes += 4;
             PRINTF ("type=ILU,sbh=\"%*.*s\"", sbhlen, sbhlen, profile_bytes);
           }
           break;

      default:
           PRINTF ("type=%lu,bytes=%lu", i, profile_tag, profile_len);
           break;
    }
    PUTCHAR (')');
  }
  return (ior - orig_ior);
}

static u_int print_context (u_char *msg, u_int little_endian)
{
  u_int32_t context_id, context_len;
  u_char   *orig_msg_ptr = msg;

  context_id  = gtohl (*(u_int32_t *) msg, little_endian);
  msg        += 4;
  context_len = gtohl (*(u_int32_t *) msg, little_endian);
  msg        += 4;

  switch (context_id)
  {
    case GIOP_SERVICECONTEXT_TRANSACTION_SERVICE:
         {
           u_int     context_endianness, context_major, context_minor;
           u_int32_t timeout, tid_len, bequal_length, otid_formatid;
           u_char   *p;
           char     *s;

           p = msg;
           context_endianness = (p[0] == 1);
           context_major = p[1];
           context_minor = p[2];
           p += 4;
           timeout = gtohl (*(u_int32_t *) p, context_endianness);
           p += 4;

           PRINTF (",context=TransactionService(timeout=%lu,current.coordinator=(",
                   timeout);
           p += print_ior (p, context_endianness);
           PUTS ("),current.terminator=(");
           p += print_ior (p, context_endianness);
           otid_formatid = gtohl (*(u_int32_t*)p, context_endianness);
           p += 4;
           bequal_length = gtohl (*(u_int32_t*)p, context_endianness);
           p += 4;
           tid_len = gtohl (*(u_int32_t *) p, context_endianness);
           p += 4;
           s = quotechars (p, tid_len);
           PRINTF ("),otid=(formatID=%lu,bequal_length=%lu,tid=<%*.*s>)",
                   otid_formatid, bequal_length, tid_len, tid_len, s);
           free (s);

           msg += ((context_len + 3) / 4) * 4;
         }
         break;

    case GIOP_SERVICECONTEXT_CODESETS:
         {
           u_int32_t short_char_codeset, char_codeset;
           u_int     context_endianness, context_major, context_minor;

           context_endianness = (msg[0] == 1);
           context_major = msg[1];
           context_minor = msg[2];
           msg += 4;
           short_char_codeset = gtohl (*(u_int32_t*)msg, context_endianness);
           msg += 4;
           char_codeset = gtohl (*(u_int32_t*)msg, context_endianness);
           msg += 4;
           PRINTF (",context=CodeSets(char=0x%x%s,wchar=0x%x%s)",
                   short_char_codeset, figure_charset_name (short_char_codeset),
                   char_codeset, figure_charset_name (char_codeset));
           context_len -= 12;
           if (context_len > 0)
           {
             PUTS ("[junk]");
             msg += (((context_len + 3) / 4) * 4);
           }
         }
         break;

    default:
         msg += ((context_len + 3) / 4) * 4;
         PRINTF (",context=%u(bytes=%u)", context_id, context_len);
         break;
  }
  return (msg - orig_msg_ptr);
}


void giop_print (u_int connection, const u_char *msg, u_int msglen)
{
  struct giop_state *s = find_state (connection);

  u_int32_t msgtype;
  u_int32_t length;
  u_int     little_endian, last_chunk, i;
  u_int     major, minor, flags;

  PRINTF ("%s <giop", Pflag ? "\n " : "");

  if (strncmp (msg, "GIOP", 4))
  {
    PUTS ("[bad magic!]\n");
    return;
  }

  if (!s)
      s = add_state (connection);

  major = msg[4];
  minor = msg[5];
  flags = msg[6];
  little_endian = ((flags & 0x1) == 0x1);
  last_chunk    = ((minor > 0) ? ((flags & 0x2) == 0) : 1);

  msgtype = msg[7];
  msg += 8;

  length = gtohl (*((u_int32_t *) msg), little_endian);
  msg += 4;

  switch (msgtype)
  {
    case GIOP_MSGTYPE_REQUEST:
         {
           u_int     response_expected;
           u_int32_t contexts, sn;
           u_int32_t okeylen, mnamelen;
           char     *okey, *mname;

           PRINTF (" i(vers=%u.%u", major, minor);

           contexts = gtohl (*(u_int32_t*)msg, little_endian);
           msg += 4;
           for (i = 0; i < contexts; i++)
              msg += print_context ((u_char*)msg, little_endian);

           sn = gtohl (*(u_int32_t*)msg, little_endian);
           msg += 4;
           response_expected = (msg[0] != 0);
           msg += 4;
           okeylen = gtohl (*(u_int32_t*)msg, little_endian);
           msg += 4;
           okey = quotechars ((u_char*)msg, okeylen);
           msg += ((okeylen + 3) / 4) * 4;
           mnamelen = gtohl (*(u_int32_t*)msg, little_endian);
           msg += 4;
           mname = quotechars ((u_char*)msg, mnamelen - 1);
           msg += ((mnamelen + 3) / 4) * 4;
           PRINTF (",sn=%lu,method=\"%s\",disc=\"%s\")", sn, mname, okey);
           free (mname);
           free (okey);
         }
         break;   

    case GIOP_MSGTYPE_CANCEL_REQUEST:
         {
           u_int32_t req_to_cancel = gtohl (*(u_int32_t*)msg, little_endian);

           msg += 4;
           PRINTF (" c(CancelRequest,vers=%u.%u,req=%lu)",
                   major, minor, req_to_cancel);
         }
         break;

    case GIOP_MSGTYPE_LOCATE_REQUEST:
         {
           u_int32_t sn, okeylen;

           sn = gtohl (*(u_int32_t*)msg, little_endian);
           msg += 4;
           okeylen = gtohl (*(u_int32_t*)msg, little_endian);
           msg += 4;
           msg += okeylen;
           PRINTF (" c(LocateRequest,vers=%u.%u,sn=%lu,bytes=%lu)",
                   major, minor, sn);
         }
         break;

    case GIOP_MSGTYPE_LOCATE_REPLY:
         {
           u_int32_t sn, repstat;

           sn = gtohl (*(u_int32_t*)msg, little_endian);
           msg += 4;
           repstat = gtohl (*(u_int32_t*)msg, little_endian);
           msg += 4;
           PRINTF (" c(LocateReply,vers=%u.%u,sn=%lu", major, minor, sn);

           switch (repstat)
           {
             case IIOP_OBJECT_FORWARD:
                  PUTS ("status=ObjectForward(");
                  msg += print_ior ((u_char *) msg, little_endian);
                  PUTCHAR (')');
                  break;
             case IIOP_OBJECT_HERE:
                  PUTS ("status=ObjectHere");
                  break;
             case IIOP_UNKNOWN_OBJECT:
                  PUTS ("status=UnknownObject");
                  break;
             default:
                  PRINTF ("[unknown reply status %lu!]\n", repstat);
                  return;
           }
           PUTCHAR (')');
         }
         break;

    case GIOP_MSGTYPE_MESSAGE_ERROR:
         PRINTF (" c(MessageError,vers=%u.%u)", major, minor);
         break;

    case GIOP_MSGTYPE_CLOSE_CONNECTION:
         PRINTF (" c(CloseConnection,vers=%u.%u)", major, minor);
         break;

    case GIOP_MSGTYPE_REPLY:
         {
           u_int32_t contexts, sn;
           u_int32_t elen, minor_code, completed, status;
           u_char   *eid;

           PRINTF (" r(vers=%u.%u", major, minor);

           contexts = gtohl (*(u_int32_t*)msg, little_endian);
           msg += 4;
           for (i = 0; i < contexts; i++)
              msg += print_context ((u_char*)msg, little_endian);

           sn = gtohl (*(u_int32_t*)msg, little_endian);
           msg += 4;
           PRINTF (",sn=%lu", sn);
           status = gtohl (*(u_int32_t*)msg, little_endian);
           msg += 4;
           switch (status)
           {
             case GIOP_REPLYSTATUS_NO_EXCEPTION:
                  PUTS (",Success");
                  break;

             case GIOP_REPLYSTATUS_USER_EXCEPTION:
                  elen = gtohl (*(u_int32_t*)msg, little_endian);
                  msg += 4;
                  eid = (u_char *) msg;
                  msg += ((elen + 3) / 4) * 4;
                  PRINTF (",UserException=\"%*.*s\"", elen - 1, elen - 1, eid);
                  break;

             case GIOP_REPLYSTATUS_SYSTEM_EXCEPTION:
                  elen = gtohl (*(u_int32_t*)msg, little_endian);
                  msg += 4;
                  eid = (u_char *) msg;
                  msg += ((elen + 3) / 4) * 4;
                  minor_code = gtohl (*(u_int32_t*)msg, little_endian);
                  msg += 4;
                  completed = gtohl (*(u_int32_t*)msg, little_endian);
                  msg += 4;
                  PRINTF (",SystemException=\"%*.*s\",minor=%lu,completed=%s",
                          elen - 1, elen - 1, eid, minor_code,
                          completed == 0 ? "Yes" : ((completed == 1) ? "No" : "Maybe"));
                  break;

             case GIOP_REPLYSTATUS_LOCATION_FORWARD:
                  {
                    PUTS (",LocationForward(");
                    msg += print_ior ((u_char*)msg, little_endian);
                    PUTCHAR (')');
                  }
                  break;

             default:
                  PRINTF ("[bad reply status %lu!]\n", status);
                  return;
           }
           PUTCHAR (')');
         }
         break;

    case GIOP_MSGTYPE_FRAGMENT:
         PRINTF (" c(Fragment,vers=%u.%u)", major, minor);
         break;

    default:
         PRINTF ("[unknown msg type %lu!]\n", msgtype);
         return;
  }
  PUTCHAR ('>');
}
