/*
 * Copyright (c) 1988, 1989, 1990, 1991, 1993, 1994
 * The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#if 0
static const char rcsid[] = "@(#) $Header: /tcpdump/master/tcpdump/print-icmp6.c,v 1.42 2000/12/13 07:57:05 itojun Exp $";
#endif

#include <stdio.h>
#include <string.h>

#include <sys/param.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>

#include "interfac.h"
#include "a2name.h"
#include "ip6.h"
#include "udp.h"
#include "ah.h"
#include "icmp6.h"

#ifdef USE_INET6

#ifndef ICMPV6_ROUTER_SOLICITATION
#define ICMPV6_ROUTER_SOLICITATION     133
#endif

#ifndef ICMPV6_ROUTER_ADVERTISEMENT
#define ICMPV6_ROUTER_ADVERTISEMENT    134
#endif

#ifndef ICMPV6_NEIGHBOR_SOLICITATION
#define ICMPV6_NEIGHBOR_SOLICITATION   135
#endif

#ifndef ICMPV6_NEIGHBOR_ADVERTISEMENT
#define ICMPV6_NEIGHBOR_ADVERTISEMENT  136
#endif

#define ICMPV6_MINLEN   8

/* Error messages and codes.
 */
#define ICMPV6_UNREACH               1    /* dest unreachable, codes: */
#define   ICMPV6_UNREACH_NOROUTE     0    /* No route to dest. */
#define   ICMPV6_UNREACH_ADMIN       1    /* Admin. prohibited */
#define   ICMPV6_UNREACH_NOTNEIGHBOR 2    /* For strict source routing. */
#define   ICMPV6_UNREACH_ADDRESS     3    /* Address unreach. */
#define   ICMPV6_UNREACH_PORT        4    /* Port unreachable */
#define ICMPV6_TOOBIG                2    /* Packet too big. */
#define ICMPV6_TIMXCEED              3    /* time exceeded, code: */
#define   ICMPV6_TIMXCEED_INTRANS    0    /* ttl==0 in transit */
#define   ICMPV6_TIMXCEED_REASS      1    /* Reassembly t.o. */
#define ICMPV6_PARAMPROB             4    /* ip header bad */
#define   ICMPV6_PARAMPROB_PROB      0    /* Actual incorrect parameter. */
#define   ICMPV6_PARAMPROB_NEXTHDR   1    /* Bad next hdr. */
#define   ICMPV6_PARAMPROB_BADOPT    2    /* Unrec. option */

/* Info messages.
 */
#define ICMPV6_ECHO        128     /* echo service */
#define ICMPV6_ECHOREPLY   129     /* echo reply */
#define ICMPV6_GRPQUERY    130     /* Query group membership. */
#define ICMPV6_GRPREPORT   131     /* Join mcast group. */
#define ICMPV6_GRPTERM     132     /* Leave mcast group. */

#define ICMPV6_ROUTERSOL   133     /* Router solicit. */
#define ICMPV6_ROUTERADV   134     /* Router advertisement. */
#define ICMPV6_NEIGHBORSOL 135     /* Neighbor solicit. */
#define ICMPV6_NEIGHBORADV 136     /* Neighbor advertisement. */

#define ICMPV6_REDIRECT    137     /* ICMPv6 redirect. */

/* Defined this way to save some HTONL cycles on little-endian boxes.
 */
#if BYTE_ORDER == BIG_ENDIAN
  #define ICMPV6_NEIGHBORADV_RTR       0x80000000  /* Router flag. */
  #define ICMPV6_NEIGHBORADV_SOL       0x40000000  /* Solicited flag. */
  #define ICMPV6_NEIGHBORADV_OVERRIDE  0x20000000  /* Override flag. */
#else  /* BYTE_ORDER == LITTLE_ENDIAN */
  #define ICMPV6_NEIGHBORADV_RTR       0x80        /* Router flag. */
  #define ICMPV6_NEIGHBORADV_SOL       0x40        /* Solicited flag. */
  #define ICMPV6_NEIGHBORADV_OVERRIDE  0x20        /* Override flag. */
#endif

#define ICMPV6_MAXTYPE        137
#define ICMPV6_INFOTYPE(type) ((type) >= 128)


struct na_hdr {
       u_int8_t  type;
       u_int8_t  code;
       u_int16_t cksum;
       u_int32_t flags;
#define NA_ROUTER  htonl(0x80000000)
#define NA_SOLICITED   htonl(0x40000000)
#define NA_OVERRIDE    htonl(0x20000000)
#define NA_RESERVED    htonl(0x1FFFFFFF)
       struct in6_addr targ_addr;
     };

struct ns_hdr {
       u_int8_t        type;
       u_int8_t        code;
       u_int16_t       cksum;
       u_int32_t       reserved;
       struct in6_addr target;
     };

struct red_hdr {
       u_int8_t        type;
       u_int8_t        code;
       u_int16_t       cksum;
       u_int32_t       reserved;
       struct in6_addr target;
       struct in6_addr destination;
     };

struct rs_hdr {
       u_int8_t  type;
       u_int8_t  code;
       u_int16_t cksum;
       u_int32_t reserved;
     };

struct ra_hdr {
       u_int8_t  type;
       u_int8_t  code;
       u_int16_t cksum;
       u_int8_t  hoplimit;
       u_int8_t  flags;
#define RA_MANAGED         0x80
#define RA_OTHER_STATEFUL  0x40
#define RA_RESERVED        0x3F
       u_int16_t router_lft;
       u_int32_t reachable;
       u_int32_t retrans;
     };

struct ra_pinfo {
       u_int8_t            type;
       u_int8_t            length;
       u_int8_t            prefix_len;
       u_int8_t            flags;
#define RA_PINFO_ONLINK    0x80
#define RA_PINFO_AUTO      0x40
#define RA_PINFO_RESERVED1 0x20
#define RA_PINFO_SITE      0x10
#define RA_PINFO_RESERVED  0x0F
       u_int32_t           valid;
       u_int32_t           prefered;
       u_int8_t            site_plen;
       u_int8_t            reserved1;
       u_int16_t           reserved2;
       struct in6_addr     prefix;
      };

struct ndisc_opt {
       u_int8_t type;
       u_int8_t length;
     };

void ndisc_opt_print (struct ndisc_opt *opt, int len)
{
  int i;

  switch (opt->type)
  {
    case 1:
         PUTS (" src[");
         break;

    case 2:
         PUTS (" target[");
         break;

    case 3:
         PUTS (" prefix[");
         if (opt->length * 8 < sizeof(struct ra_pinfo))
         {
           PUTS ("!too short]");
           return;
         }
         if (opt->length * 8 > len)
         {
           PRINTF ("%s]", ((u_char*)opt) + opt->length * 8 > snapend ?
                           "|" : "!truncated");
           return;
         }
         if (vflag > 1)
         {
           struct ra_pinfo *p = (struct ra_pinfo*)opt;

           PRINTF (" %s/%d", ip6addr_string(&p->prefix), p->prefix_len);
           if (p->flags)
           {
             PUTS (" (");
             if (p->flags & RA_PINFO_AUTO)
                PUTCHAR ('A');
             if (p->flags & RA_PINFO_ONLINK)
                PUTCHAR ('L');
             if (p->flags & RA_PINFO_RESERVED1)
                PUTCHAR ('R');
             if (p->flags & RA_PINFO_SITE)
                PUTCHAR ('S');
             if (p->flags & RA_PINFO_RESERVED)
                PRINTF ("?{%02x}", p->flags & RA_PINFO_RESERVED);
             PUTCHAR (')');
           }
           if (p->site_plen)
                PRINTF (" site %d", p->site_plen);
           if (p->reserved1)
                PRINTF (" res1 %02x", p->reserved1);
           if (p->reserved2)
                PRINTF (" res2 %04x", ntohs (p->reserved2));
           if (p->valid == 0xFFFFFFFF)
                PUTS (" valid forever");
           else if (p->valid == 0)
                PUTS (" invalid");
           else PRINTF (" valid %us", ntohl (p->valid));

           if (p->prefered == 0xFFFFFFFF)
                PUTS (" prefered forever");
           else if (p->prefered == 0)
                PUTS (" deprecated");
           else PRINTF (" pref %us", ntohl (p->prefered));
         }
         PUTCHAR (']');
         return;

    case 4:
         PUTS (" hdr[");
         if (!opt->length)
         {
           PUTCHAR (']');
           return;
         }
         if (opt->length * 8 > len)
         {
           if (((u_char *) opt) + opt->length * 8 <= snapend)
                PUTS ("!truncated");
           else PUTCHAR ('|');
           if (len <= 8)
           {
             PUTCHAR (']');
             return;
           }
           else
             PUTS (": ");
           len -= 8;
         }
         else
           len = (opt->length - 1) * 8;

         if (vflag > 1)
         {
           partial_frame++;
           ip6_print (((u_char *) opt) + 8, len);
           partial_frame--;
         }
         PUTCHAR (']');
         return;

    case 5:
         if (opt->length * 8 > len)
         {
           PRINTF ("%s]", ((u_char*)opt) + opt->length * 8 > snapend ?
                   "|" : "!truncated");
           return;
         }
         if (!opt->length)
         {
           PUTS ("mtu[]");
           return;
         }
         PRINTF (" mtu[%d]", ntohl(*(((u_int32_t*)opt) + 1)));
         return;

    default:
         PRINTF (" opt%d/%d", opt->type, opt->length);
         return;
  }

  if (!opt->length)
  {
    PUTCHAR (']');
    return;
  }
  if (opt->length * 8 > len)
  {
    PRINTF ("%s]", ((u_char*)opt) + opt->length * 8 > snapend ?
            "|" : "t!runcated");
    return;
  }

  for (i = 0; i < (opt->length * 8) - 2; i++)
  {
    char buf[16];

    sprintf (buf, "%02x", ((u_char*)(opt + 1))[i]);
    if (i == (opt->length * 8) - 3)
         PRINTF ("%s]", (char*)buf);
    else PRINTF ("%s:", (char*)buf);
  }
}

void ndisc_print_options (u_char *bp, int length)
{
  struct ndisc_opt *opt = (struct ndisc_opt*)bp;
  int    offset = 0;

  while (offset + 2 <= length)
  {
    ndisc_opt_print (opt, length - offset);
    if (opt->length == 0)
    {
      PUTS ("[optlen==0!]");
      break;
    }
    offset += opt->length * 8;
    opt = (struct ndisc_opt*) (bp + offset);
  }
}

void icmp6_print (const u_char *bp, u_int length)
{
  const struct icmp6_hdr *dp;
  char  buf[256], *str;
  int   left, rep_off = -1;

  dp   = (struct icmp6_hdr*)bp;
  left = (u_long) snapend - (u_long) bp;
  str  = buf;
  *str = '\0';

  if (left < ICMPV6_MINLEN)
     goto trunc;

  switch (dp->icmp6_type)
  {
    case ICMPV6_ECHOREPLY:
         str = "echo reply";
         break;

    case ICMPV6_UNREACH:
         switch (dp->icmp6_code)
         {
           case ICMPV6_UNREACH_NOROUTE:
                str = "no route";
                break;
           case ICMPV6_UNREACH_ADMIN:
                str = "admin prohibited";
                break;
           case ICMPV6_UNREACH_NOTNEIGHBOR:
                str = "not a neighbor";
                break;
           case ICMPV6_UNREACH_ADDRESS:
                str = "host unreachable";
                break;
           case ICMPV6_UNREACH_PORT:
                str = "port unreachable";
                break;
         }
         rep_off = 8;
         break;

    case ICMPV6_TOOBIG:
         sprintf (str, "packet too big: mtu[%lu]", ntohl (dp->icmp6_mtu));
         rep_off = 8;
         break;

    case ICMPV6_ECHO:
         str = "echo request";
         break;

    case ICMPV6_GRPQUERY:
#if 0 /* !! */
         if (!memcmp (dp+1, "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0", 16))
              sprintf (str, "MLD general query: resptime=%dmsec",
                       ntohs (dp->icmp_grpdelay));
         else sprintf (str, "MLD query: %s resptime=%dmsec",
                       ip6addr_string ((struct in6_addr*) (dp + 1)),
                       ntohs (dp->icmp_grpdelay));
#else
         strcpy (str, "MLD query");
#endif
         break;

    case ICMPV6_GRPREPORT:
         sprintf (str, "MLD report: %s",
                  ip6addr_string ((struct in6_addr*)(dp + 1)));
         break;

    case ICMPV6_GRPTERM:
         sprintf (str, "MLD reduction: %s",
                  ip6addr_string ((struct in6_addr*)(dp + 1)));
         break;

    case ICMPV6_TIMXCEED:
         switch (dp->icmp6_code)
         {
           case ICMPV6_TIMXCEED_INTRANS:
                str = "time exceeded in-transit";
                break;
           case ICMPV6_TIMXCEED_REASS:
                str = "reassembly time exceeded";
                break;
         }
         rep_off = 8;
         break;

    case ICMPV6_PARAMPROB:
         switch (dp->icmp6_code)
         {
           case ICMPV6_PARAMPROB_PROB:
                strcpy (str, "incorrect parameter");
                break;
           case ICMPV6_PARAMPROB_NEXTHDR:
                strcpy (str, "bad next header");
                break;
           case ICMPV6_PARAMPROB_BADOPT:
                strcpy (str, "unrec. option");
                break;
         }
         sprintf (str + strlen(str), " ptr[%lu]", ntohl(dp->icmp6_pptr));
         rep_off = 8;
         break;

    case ICMPV6_ROUTERSOL:
         {
           struct rs_hdr *rs;

           PUTS ("icmpv6: RS");
           if (left < sizeof(*rs))
           {
             PUTS ("[|]");
             return;
           }
           rs = (struct rs_hdr *) bp;
           if (rs->reserved)
              PRINTF (" ?%08x", ntohl (rs->reserved));
           if (vflag)
              ndisc_print_options ((u_char*)(rs+1), left - sizeof(*rs));
           return;
         }

    case ICMPV6_ROUTERADV:
         {
           struct ra_hdr *ra;

           PUTS ("icmpv6: RA");
           if (left < sizeof(*ra))
           {
             PUTS ("[|]");
             return;
           }
           ra = (struct ra_hdr*)bp;
           if (ra->flags)
           {
             PUTCHAR ('(');
             if (ra->flags & RA_MANAGED)
                PUTCHAR ('M');
             if (ra->flags & RA_OTHER_STATEFUL)
                PUTCHAR ('O');
             if (ra->flags & RA_RESERVED)
                PRINTF ("?%02x", ra->flags & RA_RESERVED);
             PUTCHAR (')');
           }
           if (vflag)
           {
             if (ra->router_lft)
                PRINTF (" default %ds", ntohs (ra->router_lft));
             if (ra->hoplimit)
                PRINTF (" hoplimit %d", ra->hoplimit);
             if (ra->reachable)
                PRINTF (" reachable %dms", ntohl (ra->reachable));
             if (ra->retrans)
                PRINTF (" retrans %dms", ntohl (ra->retrans));
             ndisc_print_options ((u_char*)(ra+1), left - sizeof(*ra));
           }
           return;
         }

    case ICMPV6_NEIGHBORSOL:
         {
           struct ns_hdr *ns;

           PUTS ("icmpv6: NS");
           if (left < sizeof(*ns))
           {
             PUTS ("[|]");
             return;
           }
           ns = (struct ns_hdr*)bp;
           PRINTF (" %s", ip6addr_string (&ns->target));
           if (ns->reserved)
              PRINTF (" ?%08x", ntohl (ns->reserved));
           if (vflag)
              ndisc_print_options ((u_char*)(ns+1), left - sizeof(*ns));
           return;
         }

    case ICMPV6_NEIGHBORADV:
         {
           struct na_hdr *na;

           PUTS ("icmpv6: NA");
           if (left < sizeof(*na))
           {
             PUTS ("[|]");
             return;
           }
           na = (struct na_hdr*)bp;

                   PRINTF (" %s", ip6addr_string (&na->targ_addr));
           if (na->flags)
           {
             PUTCHAR ('(');
             if (na->flags & NA_ROUTER)
                PUTCHAR ('R');
             if (na->flags & NA_SOLICITED)
                PUTCHAR ('S');
             if (na->flags & NA_OVERRIDE)
                PUTCHAR ('O');
             if (na->flags & NA_RESERVED)
                PRINTF ("?%08x", ntohl (na->flags & NA_RESERVED));
             PUTCHAR (')');
           }
           if (vflag)
              ndisc_print_options ((u_char*)(na+1), left - sizeof(*na));
           return;
         }

    case ICMPV6_REDIRECT:
         {
           struct red_hdr *red;

           PUTS ("icmpv6: redirect ");
           if (left < sizeof(*red))
           {
             PUTS ("[|]");
             return;
           }
           red = (struct red_hdr*)bp;

           PRINTF (" %s", ip6addr_string (&red->destination));
           if (memcmp (&red->destination, &red->target, 16) == 0)
                PUTS (" direct");
           else PRINTF (" to %s", ip6addr_string (&red->target));
           if (vflag)
             ndisc_print_options ((u_char*)(red + 1), left - sizeof(*red));
           return;
         }

    default:
         sprintf (buf, "type#%d/%d", dp->icmp6_type, dp->icmp6_code);
         break;
  }

  PRINTF ("icmpv6: %s", str);
  if (vflag > 1 && rep_off > 0)
  {
    PUTS (" Offending pkt: ");
    partial_frame++;
    ip6_print (bp + rep_off, length - rep_off);
    partial_frame--;
  }
  return;

trunc:
  PUTS ("[|icmpv6]");
}

#endif
