/*
 * Copyright (c) 1989, 1990, 1991, 1993, 1994
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#if 0
static const char rcsid[] =
"@(#) $Header: /tcpdump/master/tcpdump/print-ripng.c,v 1.2.2.1 2000/01/11 06:58:26 fenner Exp $";
#endif

#include <errno.h>
#include <stdio.h>

#include <sys/param.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>
#include <netinet/udp.h>
#include <netinet/udp_var.h>
#include <netinet/ipv6.h>

#include "route6d.h"
#include "interfac.h"
#include "a2name.h"

#ifdef USE_INET6

#ifndef IN6_IS_ADDR_UNSPECIFIED
  #define IN6_IS_ADDR_UNSPECIFIED(a)  \
          ((*(u_int32_t *)(&(a)->s6_addr[0]) == 0) && \
           (*(u_int32_t *)(&(a)->s6_addr[4]) == 0) && \
           (*(u_int32_t *)(&(a)->s6_addr[8]) == 0) && \
           (*(u_int32_t *)(&(a)->s6_addr[12]) == 0))
#endif

static int rip6_entry_print (const struct netinfo6 *ni, int metric)
{
  int l = PRINTF ("%s/%d", ip6addr_string (&ni->rip6_dest), ni->rip6_plen);

  if (ni->rip6_tag)
     l += PRINTF (" [%d]", ntohs (ni->rip6_tag));
  if (metric)
     l += PRINTF (" (%d)", ni->rip6_metric);
  return l;
}

void ripng_print (const u_char * dat, u_int length)
{
  const struct rip6     *rp = (struct rip6 *) dat;
  const struct netinfo6 *ni;
  int   amt = snapend - dat;
  int   trunc, j;
  int   i = min (length, amt) - (sizeof(*rp) - sizeof(*ni));

  if (i < 0)
     return;

  switch (rp->rip6_cmd)
  {
    case RIP6_REQUEST:
         j = length / sizeof(*ni);
         if (j == 1 &&
             rp->rip6_nets->rip6_metric == HOPCNT_INFINITY6 &&
             IN6_IS_ADDR_UNSPECIFIED (&rp->rip6_nets->rip6_dest))
         {
           PUTS (" ripng-req dump");
           break;
         }
         if (j * sizeof(*ni) != length - 4)
              PRINTF (" ripng-req %d[%d]:", j, length);
         else PRINTF (" ripng-req %d:", j);
         trunc = ((i / sizeof(*ni)) * sizeof(*ni) != i);
         for (ni = rp->rip6_nets; (i -= sizeof(*ni)) >= 0; ++ni)
         {
           if (vflag)
                PUTS ("\n\t");
           else PUTCHAR (' ');
           rip6_entry_print (ni, 0);
         }
         break;

    case RIP6_RESPONSE:
         j = length / sizeof(*ni);
         if (j * sizeof(*ni) != length - 4)
              PRINTF (" ripng-resp %d[%d]:", j, length);
         else PRINTF (" ripng-resp %d:", j);
         trunc = ((i / sizeof(*ni)) * sizeof(*ni) != i);
         for (ni = rp->rip6_nets; (i -= sizeof(*ni)) >= 0; ++ni)
         {
           if (vflag)
                PUTS ("\n\t");
           else PUTCHAR (' ');
           rip6_entry_print (ni, ni->rip6_metric);
         }
         if (trunc)
            PUTS ("[|rip]");
         break;

    default:
         PRINTF (" ripng-%d ?? %d", rp->rip6_cmd, length);
         break;
  }
  if (rp->rip6_vers != RIP6_VERSION)
     PRINTF (" [vers %d]", rp->rip6_vers);
}
#endif     /* USE_INET6 */
