/*
 * Copyright (c) 1988, 1989, 1990, 1991, 1993, 1994
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#if 0
static const char rcsid[] =
"@(#) $Header: /tcpdump/master/tcpdump/print-rt6.c,v 1.3.2.1 2000/01/11 06:58:26 fenner Exp $";
#endif

#include <stdio.h>

#include <sys/param.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <net/if.h>

#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/ip_var.h>
#include <netinet/udp.h>
#include <netinet/udp_var.h>
#include <netinet/tcp.h>
#include <netinet/ipv6.h>

#include "interfac.h"
#include "a2name.h"

#ifdef USE_INET6

#define ip6r0_reserved     bitmap         /* .h-file trouble */
#define ip6r0_len          rt_hdr.hdrlen  /* .h-file trouble */

#define IPV6_RTHDR_LOOSE   0 /* this hop need not be a neighbor. XXX old spec */
#define IPV6_RTHDR_STRICT  1 /* this hop must be a neighbor. XXX old spec */
#define IPV6_RTHDR_TYPE_0  0 /* IPv6 routing header type 0 */

int rt6_print (const u_char *bp, const u_char *bp2)
{
  const struct ipv6_rt_hdr  *dp  = (struct ipv6_rt_hdr*) bp;
  const struct ipv6_rt0_hdr *dp0;
  const struct ipv6hdr      *ip  = (struct ipv6hdr*) bp2;
  const u_char              *ep  = snapend;
  int   i, len = dp->hdrlen;

  PRINTF ("%s > %s: ",
          ip6addr_string (&ip->ipv6_src),
          ip6addr_string (&ip->ipv6_dst));

  TCHECK (dp->segments_left);

  PRINTF ("srcrt (len=%d, ", dp->hdrlen);
  PRINTF ("type=%d, ",       dp->type);
  PRINTF ("segleft=%d, ",    dp->segments_left);

  switch (dp->type)
  {
    case IPV6_RTHDR_TYPE_0:
         dp0 = (struct ipv6_rt0_hdr*) dp;

         TCHECK (dp0->ip6r0_reserved);
         if (dp0->ip6r0_reserved || vflag)
            PRINTF ("rsv=0x%0lx, ", ntohl (dp0->ip6r0_reserved));

         if (len % 2 == 1)
            goto trunc;
         len >>= 1;
         for (i = 0; i < len; i++)
         {
           struct in6_addr *addr = ((struct in6_addr*) (dp0 + 1)) + i;
           if ((u_char *) addr > ep - sizeof(*addr))
              goto trunc;

           PRINTF ("[%d]%s", i, ip6addr_string ((u_char*)addr));
           if (i != len - 1)
              PUTS (", ");
         }
         PUTCHAR (')');
         return ((dp0->ip6r0_len + 1) << 3);

    default:
         goto trunc;
  }

trunc:
  PUTS ("[|srcrt]");
  return 65535;
}
#endif     /* USE_INET6 */
