/* 
 * Copyright (c) 1992, 1993, 1994, 1995, 1996
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include <sys/param.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>
#include <rpc/rpc.h>
#include <rpc/pmap_pro.h>
#include <netdb.h>

#include "interfac.h"
#include "a2name.h"

static struct tok proc2str[] = {
  { PMAPPROC_NULL,   "null" },
  { PMAPPROC_SET,    "set" },
  { PMAPPROC_UNSET,  "unset" },
  { PMAPPROC_GETPORT,"getport" },
  { PMAPPROC_DUMP,   "dump" },
  { PMAPPROC_CALLIT, "call" },
  { 0,               NULL }
};

static char *progstr (u_int);

void sunrpcrequest_print (const u_char *bp, u_int length, const u_char *bp2)
{
  const struct rpc_msg *rp = (struct rpc_msg *) bp;
  const struct ip      *ip = (struct ip *) bp2;
  u_int x;

  if (!nflag)
       PRINTF ("%s.%x > %s.sunrpc: %d",
               ipaddr_string (&ip->ip_src),
               (u_int) ntohl (rp->rm_xid),
               ipaddr_string (&ip->ip_dst),
               length);
  else PRINTF ("%s.%x > %s.%x: %d",
               ipaddr_string (&ip->ip_src),
               (u_int) ntohl (rp->rm_xid),
               ipaddr_string (&ip->ip_dst),
               PMAPPORT,
               length);
  PRINTF (" %s", tok2str (proc2str, " proc #%lu",
          ntohl (rp->rm_call.cb_proc)));
  x = ntohl (rp->rm_call.cb_rpcvers);
  if (x != 2)
     PRINTF (" [rpcver %u]", x);

  switch (ntohl (rp->rm_call.cb_proc))
  {
    case PMAPPROC_SET:
    case PMAPPROC_UNSET:
    case PMAPPROC_GETPORT:
    case PMAPPROC_CALLIT:
         x = ntohl (rp->rm_call.cb_prog);
         if (!nflag)
              PRINTF (" %s", progstr (x));
         else PRINTF (" %u", x);
         PRINTF (".%u", (u_int) ntohl (rp->rm_call.cb_vers));
         break;
  }
}

static char * progstr (u_int prog)
{
  static char buf[32];
  static int  lastprog = 0;

#ifdef USE_RPC
  struct rpcent *rp;

  if (lastprog && prog == lastprog)
     return (buf);

  rp = getrpcbynumber (prog);
  if (!rp)
       sprintf (buf, "#%u", prog);
  else strcpy (buf, rp->r_name);

#else
  if (lastprog && prog == lastprog)
     return (buf);
  sprintf (buf, "#%u", prog);
#endif

  lastprog = prog;
  return (buf);
}

