/*
 * Copyright (c) 2000 Ben Smithurst <ben@scientia.demon.co.uk>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#if 0
static const char rcsid[] = "@(#) $Header: /tcpdump/master/tcpdump/print-timed.c,v 1.1 2000/10/06 05:35:37 guy Exp $";
#endif

#include <stdio.h>
#include <string.h>

#include <sys/param.h>
#include <sys/time.h>
#include <netinet/in.h>

#include "ip.h"
#include "udp.h"
#include "tcp.h"

#include "interfac.h"
#include "a2name.h"
#include "extract.h"            /* must come after interface.h */
#include "timed.h"

static char *tsptype[TSPTYPENUMBER] = {
            "ANY", "ADJTIME", "ACK", "MASTERREQ", "MASTERACK",
            "SETTIME", "MASTERUP", "SLAVEUP", "ELECTION",
            "ACCEPT", "REFUSE", "CONFLICT", "RESOLVE", "QUIT",
            "DATE", "DATEREQ", "DATEACK", "TRACEON", "TRACEOFF",
            "MSITE", "MSITEREQ", "TEST", "SETDATE", "SETDATEREQ",
            "LOOP"
          };

void timed_print (const u_char * bp, u_int length)
{
#define endof(x) ((u_char *)&(x) + sizeof(x))
  struct tsp *tsp = (struct tsp *) bp;
  long   sec, usec;
  const  u_char *end;

  if (endof (tsp->tsp_type) > snapend)
  {
    PUTS ("[|timed]");
    return;
  }
  if (tsp->tsp_type < TSPTYPENUMBER)
       PRINTF ("TSP_%s", tsptype[tsp->tsp_type]);
  else PRINTF ("(tsp_type %#x)", tsp->tsp_type);

  if (endof (tsp->tsp_vers) > snapend)
  {
    PUTS (" [|timed]");
    return;
  }

  PRINTF (" vers %d", tsp->tsp_vers);

  if (endof (tsp->tsp_seq) > snapend)
  {
    PUTS (" [|timed]");
    return;
  }
  PRINTF (" seq %d", tsp->tsp_seq);

  if (tsp->tsp_type == TSP_LOOP)
  {
    if (endof (tsp->tsp_hopcnt) > snapend)
    {
      PUTS (" [|timed]");
      return;
    }
    PRINTF (" hopcnt %d", tsp->tsp_hopcnt);
  }
  else if (tsp->tsp_type == TSP_SETTIME ||
           tsp->tsp_type == TSP_ADJTIME || tsp->tsp_type == TSP_SETDATE || tsp->tsp_type == TSP_SETDATEREQ)
  {
    if (endof (tsp->tsp_time) > snapend)
    {
      PUTS (" [|timed]");
      return;
    }
    sec  = ntohl ((long) tsp->tsp_time.tv_sec);
    usec = ntohl ((long) tsp->tsp_time.tv_usec);
    if (usec < 0)
    {
      /* corrupt, skip the rest of the packet */
      return;
    }
    PUTS (" time ");
    if (sec < 0 && usec != 0)
    {
      sec++;
      if (sec == 0)
         PUTCHAR ('-');
      usec = 1000000 - usec;
    }
    PRINTF ("%ld.%06ld", sec, usec);
  }

  end = memchr (tsp->tsp_name, '\0', snapend - (u_char*)tsp->tsp_name);
  if (end == NULL)
       PUTS (" [|timed]");
  else PRINTF (" name %*.s", end - (u_char*)tsp->tsp_name, tsp->tsp_name);
}
