/*
 * Rx protocol format
 *
 * $Id: rx.h,v 1.1 1999/11/17 05:45:58 assar Exp $
 */

#ifndef __RX_H
#define __RX_H

#define FS_RX_PORT     7000
#define CB_RX_PORT     7001
#define PROT_RX_PORT   7002
#define VLDB_RX_PORT   7003
#define KAUTH_RX_PORT  7004
#define VOL_RX_PORT    7005
#define ERROR_RX_PORT  7006   /* Doesn't seem to be used */
#define BOS_RX_PORT    7007

#ifndef AFSNAMEMAX
#define AFSNAMEMAX     256
#endif

#ifndef AFSOPAQUEMAX
#define AFSOPAQUEMAX   1024
#endif

#define PRNAMEMAX      64
#define VLNAMEMAX      65
#define KANAMEMAX      64
#define BOSNAMEMAX     256

#define PRSFS_READ        1   /* Read files */
#define PRSFS_WRITE       2   /* Write files */
#define PRSFS_INSERT      4   /* Insert files into a directory */
#define PRSFS_LOOKUP      8   /* Lookup files into a directory */
#define PRSFS_DELETE      16  /* Delete files */
#define PRSFS_LOCK        32  /* Lock files */
#define PRSFS_ADMINISTER  64  /* Change ACL's */

#ifdef OLD_RX_PRINTER
struct rx_header {
       u_int32_t epoch;
       u_int32_t cid;
       u_int32_t callNumber;
       u_int32_t seq;
       u_int32_t serial;
       u_char    type;
#define RX_PACKET_TYPE_DATA      1
#define RX_PACKET_TYPE_ACK       2
#define RX_PACKET_TYPE_BUSY      3
#define RX_PACKET_TYPE_ABORT     4
#define RX_PACKET_TYPE_ACKALL    5
#define RX_PACKET_TYPE_CHALLENGE 6
#define RX_PACKET_TYPE_RESPONSE  7
#define RX_PACKET_TYPE_DEBUG     8
#define RX_PACKET_TYPE_PARAMS    9
#define RX_PACKET_TYPE_VERSION   13
       u_char flags;
#define RX_CLIENT_INITIATED      1
#define RX_REQUEST_ACK           2
#define RX_LAST_PACKET           4
#define RX_MORE_PACKETS          8
#define RX_FREE_PACKET           16
       u_char  userStatus;
       u_char  securityIndex;
       u_short spare;          /* How clever: even though the AFS */
       u_short serviceId;      /* header files indicate that the */
     };                        /* serviceId is first, it's really */
                               /* encoded _after_ the spare field */
#endif


#define NUM_RX_FLAGS 5


/* $Header: /afs/stacken.kth.se/src/SourceRepository/tcpdump/rx.h,v 1.2 1998/07/19 18:04:44 mattiasa Exp $ */

/*
****************************************************************************
*        Copyright IBM Corporation 1988, 1989 - All Rights Reserved        *
*                                                                          *
* Permission to use, copy, modify, and distribute this software and its    *
* documentation for any purpose and without fee is hereby granted,         *
* provided that the above copyright notice appear in all copies and        *
* that both that copyright notice and this permission notice appear in     *
* supporting documentation, and that the name of IBM not be used in        *
* advertising or publicity pertaining to distribution of the software      *
* without specific, written prior permission.                              *
*                                                                          *
* IBM DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL *
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL IBM *
* BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY      *
* DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER  *
* IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING   *
* OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.    *
****************************************************************************
*/


/* This is the address of the data portion of the packet.  Any encryption
 * headers will be at this address, the actual data, for a data packet, will
 * start at this address + the connection's security header size.
 */
#define rx_DataOf(packet)               ((char *) (packet)->wire.data)
#define rx_GetDataSize(packet)          ((packet)->length)
#define rx_SetDataSize(packet, size)    ((packet)->length = (size))

/* These macros used in conjunction with reuse of packet header spare as a
 * packet cksum for rxkad security module.
 */
#define rx_GetPacketCksum(packet)        ((packet)->header.spare)
#define rx_SetPacketCksum(packet, cksum) ((packet)->header.spare = (cksum))

/* The offset of the actual user's data in the packet, skipping any
 * security header
 */
#define rx_UserDataOf(conn, packet) (((char*) (packet)->wire.data) + (conn)->securityHeaderSize)

/* This is the maximum size data packet that can be sent on this
 * connection, accounting for security module-specific overheads.
 */
#define rx_MaxUserDataSize(conn) ((conn)->peer->packetSize - RX_HEADER_SIZE - \
                                  (conn)->securityHeaderSize - (conn)->securityMaxTrailerSize)

struct rx_securityObjectStats {
       char      type;           /* 0:unk 1:null,2:vab 3:kad */
       char      level;
       char      sparec[10];     /* force correct alignment */
       int32_t   flags;          /* 1=>unalloc, 2=>auth, 4=>expired */
       u_int32_t expires;
       u_int32_t packetsReceived;
       u_int32_t packetsSent;
       u_int32_t bytesReceived;
       u_int32_t bytesSent;
       short     spares[4];
       int32_t   sparel[8];
     };

/* Bottom n-bits of the Call Identifier give the call number
 */
#define RX_MAXCALLS     4   /* Power of 2; max async calls per connection */
#define RX_CIDSHIFT     2   /* Log2(RX_MAXCALLS) */
#define RX_CHANNELMASK  (RX_MAXCALLS-1)
#define RX_CIDMASK      (~RX_CHANNELMASK)


#ifndef OLD_RX_PRINTER

/* The rx part of the header of a packet, in host form
 */
struct rx_header {
       u_int32_t epoch;      /* Start time of client process */
       u_int32_t cid;        /* Connection id (defined by client) */
       u_int32_t callNumber; /* Current call number */
       u_int32_t seq;        /* Sequence number of this packet, within this call */
       u_int32_t serial;     /* Serial number of this packet: a new serial
                              * number is stamped on each packet sent out */
       u_char    type;       /* RX packet type */
       u_char    flags;      /* Flags, defined below */
       u_char    userStatus; /* User defined status information, returned/set
                              * by macros rx_Get/SetLocal/RemoteStatus */
       u_char    securityIndex; /* Which service-defined security method to use */
       u_short   serviceId;     /* The id of the server-provided service that
                                 * this packet is directed to */

       /* This spare is now used for packet header checkksum.  see
        * rxi_ReceiveDataPacket and packet cksum macros above for details.
        */
       u_short spare;
     };
#endif


/* Flags for rx_header flags field
 */
#define RX_CLIENT_INITIATED 1   /* Packet is sent/received from client side of call */
#define RX_REQUEST_ACK      2   /* Peer requests acknowledgement */
#define RX_LAST_PACKET      4   /* This is the last packet from this side of the call */
#define RX_MORE_PACKETS     8   /* There are more packets following
                                 * this, i.e. the next sequence number
                                 * seen by the receiver should be greater
                                 * than this one, rather than a retransmission
                                 * of an earlier sequence number */

/* The following flags are preset per packet, i.e. they don't change on
 * retran smission of the packet
 */
#define RX_PRESET_FLAGS  (RX_CLIENT_INITIATED | RX_LAST_PACKET)

#define RX_FLAGS { "....", "C...", ".A..", "CA..", "..L.", "C.L.", ".AL.", \
                   "CAL.", "...M", "C..M", ".A.M", "CA.M", "..LM", "C.LM", \
                   ".ALM", "CALM" }

#define RX_IPUDP_SIZE  28

/* to use these larger packet sizes, we probably should revise the
 * packet struct so that the data buffer isn't statically defined in
 * size, but the max MTU that this system can support is determined at
 * startup, and the packet structures all use that instead of (eg)
 * 4352, of which 3/4 would be wasted 98% of the time.  So all these
 * SIZEs become set-once variables.  */

/* REMOTE_PACKET_SIZE is currently the same as local.  This is because REMOTE
 * is defined much too generally for my tastes, and includes the case of
 * multiple class C nets connected with a router within one campus or MAN.
 * I don't want to make local performance suffer just because of some
 * out-dated protocol that used to be in use on the NSFANET that's
 * practically unused anymore.  Any modern IP implementation will be
 * using MTU discovery, and even old routers shouldn't frag packets
 * when sending from one connected network directly to another.  Maybe
 * the next release of RX will do MTU discovery.
 */
#if 0
#define RX_MAX_DL_MTU  4352  /* what FDDI(RFC1188) uses... Larger? */
#define RX_MAX_DL_MTU 17914  /* what IBM 16MB TR  uses   */
#define RX_MAX_DL_MTU  8166  /* IEEE 802.4 */
#define RX_MAX_DL_MTU  4464  /* IEEE 802.5 MAX */
#define RX_MAX_DL_MTU  2002  /* IEEE 802.5 Recommended */
#endif

#define RX_MAX_DL_MTU           1500  /* what Ethernet uses */


#define RX_MAX_PACKET_SIZE      (RX_MAX_DL_MTU -RX_IPUDP_SIZE)
#define RX_HEADER_SIZE          sizeof(struct rx_header) /* Size of rx header */
#define RX_MAX_PACKET_DATA_SIZE (RX_MAX_PACKET_SIZE-RX_HEADER_SIZE)

#ifdef AFS_HPUX_ENV
  #define RX_LOCAL_PACKET_SIZE  (1497 - RX_IPUDP_SIZE)
  #define RX_REMOTE_PACKET_SIZE (1497 - RX_IPUDP_SIZE)
#else
  #define RX_LOCAL_PACKET_SIZE  RX_MAX_PACKET_SIZE  /* For hosts on same net */
  #define RX_REMOTE_PACKET_SIZE RX_MAX_PACKET_SIZE  /* see note above */
#endif

#if 0
#define RX_REMOTE_PACKET_SIZE   (576-RX_IPUDP_SIZE) /* For Internet hosts */
#endif

#define RX_PP_PACKET_SIZE       576                 /* tolerable SLIP */

#define RX_MINIMUM_PACKET_SIZE  (200-RX_IPUDP_SIZE)

/* Packet types, for rx_packet.type
 */
#define RX_PACKET_TYPE_DATA       1    /* A vanilla data packet */
#define RX_PACKET_TYPE_ACK        2    /* Acknowledge packet */
#define RX_PACKET_TYPE_BUSY       3    /* Busy: can't accept call immediately; try later */
#define RX_PACKET_TYPE_ABORT      4    /* Abort packet.  No response needed. */
#define RX_PACKET_TYPE_ACKALL     5    /* Acknowledges receipt of all packets */
#define RX_PACKET_TYPE_CHALLENGE  6    /* Challenge client's identity: request credentials */
#define RX_PACKET_TYPE_RESPONSE   7    /* Respond to challenge packet */
#define RX_PACKET_TYPE_DEBUG      8    /* Get debug information */
#define RX_PACKET_TYPE_PARAMS     9    /* exchange size params (showUmine) */

#define RX_PACKET_TYPES    { "data", "ack ", "busy", "abort", "ackall", \
                             "challenge", "response", "debug", "params" }
#define RX_N_PACKET_TYPES  10      /* Must agree with above list; counts 0 */

/* Packet classes, for rx_AllocPacket
 */
#define RX_PACKET_CLASS_RECEIVE   0
#define RX_PACKET_CLASS_SEND      1
#define RX_PACKET_CLASS_SPECIAL   2

#define RX_N_PACKET_CLASSES       3     /* Must agree with above list */


/* Maximum number of acknowledgements in an acknowledge packet
 */
#define RX_MAXACKS          255

/* The structure of the data portion of an acknowledge packet: An acknowledge
 * packet is in network byte order at all times.  An acknowledgement is always
 * prompted for a specific reason by a specific incoming packet.  This reason
 * is reported in "reason" and the packet's sequence number in the packet
 * header.seq.  In addition to this information, all of the current
 * acknowledgement information about this call is placed in the packet.
 * "FirstPacket" is the sequence number of the first packet represented in an
 * array of bytes, "acks", containing acknowledgement information for a number
 * of consecutive packets.  All packets prior to FirstPacket are implicitly
 * acknowledged: the sender need no longer be concerned about them.  Packets
 * from firstPacket+nAcks and on are not acknowledged.  Packets in the range
 * [firstPacket,firstPacket+nAcks) are each acknowledged explicitly.  The
 * acknowledgement may be RX_NACK if the packet is not (currently) at the
 * receiver (it may have never been received, or received and then later
 * dropped), or it may be RX_ACK if the packet is queued up waiting to be read
 * by the upper level software.  RX_ACK does not imply that the packet may not
 * be dropped before it is read; it does imply that the sender should stop
 * retransmitting the packet until notified otherwise.  The field
 * previousPacket identifies the previous packet received by the peer.  This
 * was used in a previous version of this software, and could be used in the
 * future.  The serial number in the data part of the ack packet corresponds to
 * the serial number oof the packet which prompted the acknowledge.  Any
 * packets which are explicitly not acknowledged, and which were last
 * transmitted with a serial number less than the provided serial number,
 * should be retransmitted immediately.  Actually, this is slightly inaccurate:
 * packets are not necessarily received in order.  When packets are habitually
 * transmitted out of order, this is allowed for in the retransmission
 * algorithm by introducing the notion of maximum packet skew: the degree of
 * out-of-orderness of the packets received on the wire.  This number is
 * communicated from the receiver to the sender in ack packets. */

struct rx_ackPacket {
    u_short bufferSpace;    /* Number of packet buffers available.  That is:
                             * the number of buffers that the sender of the ack
                             * packet is willing to provide for data, on this
                             * or subsequent calls.  Lying is permissable.
                             */
    u_short maxSkew;        /* Maximum difference between serial# of packet ack
                             * nowledged and highest packet yet received
                             */
    u_int32_t firstPacket;  /* The first packet in the list of acknowledged
                             * packets
                             */
    u_int32_t previousPacket; /* The previous packet number received (obsolete?) */
    u_int32_t serial;       /* Serial number of the packet which prompted the
                             * acknowledge
                             */
    u_char  reason;         /* Reason for the acknowledge of ackPacket,
                             * defined below
                             */
    u_char  nAcks;          /* Number of acknowledgements */

    u_char  acks[RX_MAXACKS]; /* Up to RX_MAXACKS packet acknowledgements,
                               * defined below
                               */
    /* Packets <firstPacket are implicitly acknowledged and may be discarded
     * by the sender.  Packets >= firstPacket+nAcks are implicitly NOT
     * acknowledged.
     * No packets with sequence numbers >= firstPacket should be discarded by
     * the sender (they may thrown out at any time by the receiver)
     */
};

/* Reason for acknowledge message
 */
#define RX_ACK_REQUESTED        1   /* Peer requested an ack on this packet */
#define RX_ACK_DUPLICATE        2   /* Duplicate packet */
#define RX_ACK_OUT_OF_SEQUENCE  3   /* Packet out of sequence */
#define RX_ACK_EXCEEDS_WINDOW   4   /* Packet sequence number higher than window; discarded */
#define RX_ACK_NOSPACE          5   /* No buffer space at all */
#define RX_ACK_PING             6   /* This is a keep-alive ack */
#define RX_ACK_PING_RESPONSE    7   /* Ack'ing because we were pinged */
#define RX_ACK_DELAY            8   /* Ack generated since nothing has happened since receiving packet */

/* Packet acknowledgement type
 */
#define RX_ACK_TYPE_NACK        0   /* I Don't have this packet */
#define RX_ACK_TYPE_ACK         1   /* I have this packet, although I may discard it later */

#define RX_ACK_REASONS " R2OWSKAD"

/* The packet size transmitted for an acknowledge is adjusted to reflect the
 * actual size of the acks array.  This macro defines the size
 */
#define rx_AckDataSize(nAcks) (sizeof(struct rx_ackPacket) - RX_MAXACKS + (nAcks))

#define RX_CHALLENGE_TIMEOUT    2   /* Number of seconds before another
                                     * authentication request packet is generated
                                     */

/* RX error codes.  RX uses error codes from -1 to -64.  Rxgen may use other
 * eeror codes < -64; user programs are expected to return positive error codes
 */

/* Something bad happened to the connection; temporary loss of communication
 */
#define RX_CALL_DEAD                (-1)

/* An invalid operation, such as a client attempting to send data after having
 * received the beginning of a reply from the server
 */
#define RX_INVALID_OPERATION        (-2)

/* An optional timeout per call may be specified
 */
#define RX_CALL_TIMEOUT             (-3)

/* End of data on a read
 */
#define RX_EOF                      (-4)

/* Some sort of low-level protocol error
 */
#define RX_PROTOCOL_ERROR           (-5)

/* Generic user abort code; used when no more specific error code needs to
 * be communicated.  For example, multi rx clients use this code to abort a
 * multi rx call
 */
#define RX_USER_ABORT               (-6)

/* Port already in use (from rx_Init)
 */
#define RX_ADDRINUSE                (-7)

#endif
